from enum import Enum
from typing import Optional

from pydantic import validator

from vocode.streaming.input_device.base_input_device import BaseInputDevice
from vocode.streaming.models.client_backend import InputAudioConfig
from vocode.streaming.telephony.constants import (
    DEFAULT_AUDIO_ENCODING,
    DEFAULT_CHUNK_SIZE,
    DEFAULT_SAMPLING_RATE,
)
from .audio_encoding import AudioEncoding
from .model import BaseModel, TypedModel


class TranscriberType(str, Enum):
    BASE = "transcriber_base"
    DEEPGRAM = "transcriber_deepgram"
    GOOGLE = "transcriber_google"
    ASSEMBLY_AI = "transcriber_assembly_ai"
    WHISPER_CPP = "transcriber_whisper_cpp"
    REV_AI = "transcriber_rev_ai"


class EndpointingType(str, Enum):
    BASE = "endpointing_base"
    TIME_BASED = "endpointing_time_based"
    PUNCTUATION_BASED = "endpointing_punctuation_based"


class EndpointingConfig(TypedModel, type=EndpointingType.BASE):
    pass


class TimeEndpointingConfig(EndpointingConfig, type=EndpointingType.TIME_BASED):
    time_cutoff_seconds: float = 0.4


class PunctuationEndpointingConfig(
    EndpointingConfig, type=EndpointingType.PUNCTUATION_BASED
):
    time_cutoff_seconds: float = 0.4


class TranscriberConfig(TypedModel, type=TranscriberType.BASE.value):
    sampling_rate: int
    audio_encoding: AudioEncoding
    chunk_size: int
    endpointing_config: Optional[EndpointingConfig] = None
    downsampling: Optional[int] = None
    min_interrupt_confidence: Optional[float] = None

    @validator("min_interrupt_confidence")
    def min_interrupt_confidence_must_be_between_0_and_1(cls, v):
        if v is not None and (v < 0 or v > 1):
            raise ValueError("must be between 0 and 1")
        return v

    @classmethod
    def from_input_device(
        cls,
        input_device: BaseInputDevice,
        endpointing_config: Optional[EndpointingConfig] = None,
        **kwargs,
    ):
        return cls(
            sampling_rate=input_device.sampling_rate,
            audio_encoding=input_device.audio_encoding,
            chunk_size=input_device.chunk_size,
            endpointing_config=endpointing_config,
            **kwargs,
        )

    @classmethod
    def from_telephone_input_device(
        cls,
        endpointing_config: Optional[EndpointingConfig] = None,
        **kwargs,
    ):
        return cls(
            sampling_rate=DEFAULT_SAMPLING_RATE,
            audio_encoding=DEFAULT_AUDIO_ENCODING,
            chunk_size=DEFAULT_CHUNK_SIZE,
            endpointing_config=endpointing_config,
            **kwargs,
        )

    @classmethod
    def from_input_audio_config(
        cls,
        input_audio_config: InputAudioConfig,
        **kwargs
    ):
        return cls(
            sampling_rate=input_audio_config.sampling_rate,
            audio_encoding=input_audio_config.audio_encoding,
            chunk_size=input_audio_config.chunk_size,
            downsampling=input_audio_config.downsampling,
            **kwargs,
        )


class DeepgramTranscriberConfig(TranscriberConfig, type=TranscriberType.DEEPGRAM.value):
    language: Optional[str] = None
    model: Optional[str] = None
    tier: Optional[str] = None
    version: Optional[str] = None
    keywords: Optional[list] = None


class GoogleTranscriberConfig(TranscriberConfig, type=TranscriberType.GOOGLE.value):
    model: Optional[str] = None


class AssemblyAITranscriberConfig(
    TranscriberConfig, type=TranscriberType.ASSEMBLY_AI.value
):
    pass


class WhisperCPPTranscriberConfig(TranscriberConfig, type=TranscriberType.WHISPER_CPP):
    buffer_size_seconds: float = 1
    libname: str
    fname_model: str


class RevAITranscriberConfig(TranscriberConfig, type=TranscriberType.REV_AI):
    pass



import io
from typing import Optional
import requests
from pydub import AudioSegment

from vocode import getenv
from vocode.streaming.synthesizer.base_synthesizer import (
    BaseSynthesizer,
    SynthesisResult,
)
from vocode.streaming.models.synthesizer import CoquiSynthesizerConfig
from vocode.streaming.agent.bot_sentiment_analyser import BotSentiment
from vocode.streaming.models.message import BaseMessage


COQUI_BASE_URL = "https://app.coqui.ai/api/v2/"


class CoquiSynthesizer(BaseSynthesizer):
    def __init__(self, config: CoquiSynthesizerConfig):
        super().__init__(config)
        self.api_key = config.api_key or getenv("COQUI_API_KEY")
        self.voice_id = config.voice_id
        self.voice_prompt = config.voice_prompt

    def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        bot_sentiment: Optional[BotSentiment] = None,
    ) -> SynthesisResult:
        url = COQUI_BASE_URL + "samples"
        if self.voice_prompt:
            url = f"{url}/from-prompt/"

        headers = {"Authorization": f"Bearer {self.api_key}"}

        emotion = "Neutral"
        if bot_sentiment.emotion:
            emotion = bot_sentiment.emotion.capitalize()

        body = {
            "text": message.text,
            "name": "unnamed",
            "emotion": emotion,
        }

        if self.voice_prompt:
            body["prompt"] = self.voice_prompt
        else:
            body["voice_id"] = self.voice_id

        # print(url, "➡️", body)
        response = requests.post(url, headers=headers, json=body)
        sample = response.json()
        # print("⬅️", sample)
        response = requests.get(sample["audio_url"])
        audio_segment: AudioSegment = AudioSegment.from_wav(
            io.BytesIO(response.content)
        )

        output_bytes: bytes = audio_segment.raw_data

        self.create_synthesis_result_from_wav(
            file=output_bytes,
            message=message,
            chunk_size=chunk_size,
        )



import io
from typing import Optional
from pydub import AudioSegment
import requests
from vocode import getenv
from vocode.turn_based.synthesizer.base_synthesizer import BaseSynthesizer

COQUI_BASE_URL = "https://app.coqui.ai/api/v2/"
DEFAULT_SPEAKER_ID = "d2bd7ccb-1b65-4005-9578-32c4e02d8ddf"


class CoquiSynthesizer(BaseSynthesizer):
    def __init__(self, voice_id: Optional[str] = None, api_key: Optional[str] = None):
        self.voice_id = voice_id or DEFAULT_SPEAKER_ID
        self.api_key = getenv("COQUI_API_KEY", api_key)

    def synthesize(self, text: str) -> AudioSegment:
        url = COQUI_BASE_URL + "samples"
        headers = {"Authorization": f"Bearer {self.api_key}"}
        body = {
            "text": text,
            "speaker_id": self.voice_id,
            "name": "unnamed",
        }
        response = requests.post(url, headers=headers, json=body)
        assert response.ok, response.text
        sample = response.json()
        response = requests.get(sample["audio_url"])
        return AudioSegment.from_wav(io.BytesIO(response.content))



import io
from typing import Optional
from pydub import AudioSegment
from vocode import getenv
from vocode.turn_based.synthesizer.base_synthesizer import BaseSynthesizer

class ElevenLabsSynthesizer(BaseSynthesizer):
    def __init__(
        self,
        voice_id: str,
        stability: Optional[float] = None,
        similarity_boost: Optional[float] = None,
        api_key: Optional[str] = None,
    ):
        import elevenlabs
        self.elevenlabs = elevenlabs

        self.voice_id = voice_id
        self.api_key = getenv("ELEVEN_LABS_API_KEY", api_key)
        self.validate_stability_and_similarity_boost(stability, similarity_boost)
        self.stability = stability
        self.similarity_boost = similarity_boost

    def validate_stability_and_similarity_boost(
        self, stability: Optional[float], similarity_boost: Optional[float]
    ) -> None:
        if (stability is None) != (similarity_boost is None):
            raise ValueError(
                "Both stability and similarity_boost must be set or not set."
            )

    def synthesize(self, text: str) -> AudioSegment:
        self.elevenlabs.set_api_key(self.api_key)
        voice = self.elevenlabs.Voice(voice_id=self.voice_id)
        if self.stability is not None and self.similarity_boost is not None:
            voice.settings = self.elevenlabs.VoiceSettings(
                stability=self.stability, similarity_boost=self.similarity_boost)

        audio = self.elevenlabs.generate(text, voice=voice)

        return AudioSegment.from_mp3(io.BytesIO(audio))



import logging
from typing import Optional
from pydub import AudioSegment
import numpy as np
import io
from vocode.streaming.agent.bot_sentiment_analyser import BotSentiment
from vocode.streaming.models.audio_encoding import AudioEncoding
from vocode.streaming.models.message import BaseMessage

from .base_synthesizer import BaseSynthesizer, SynthesisResult

from vocode.streaming.models.synthesizer import CoquiTTSSynthesizerConfig



class CoquiTTSSynthesizer(BaseSynthesizer):
    def __init__(
        self, config: CoquiTTSSynthesizerConfig, logger: Optional[logging.Logger] = None
    ):
        super().__init__(config)
        
        from TTS.api import TTS
        self.tts = TTS(**config.tts_kwargs)
        self.speaker = config.speaker
        self.language = config.language

    def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        bot_sentiment: Optional[BotSentiment] = None,
    ) -> SynthesisResult:
        
        tts = self.tts
        audio_data = np.array(tts.tts(message.text, self.speaker, self.language))

        # Convert the NumPy array to bytes
        audio_data_bytes = (audio_data * 32767).astype(np.int16).tobytes()

        # Create an in-memory file-like object (BytesIO) to store the audio data
        buffer = io.BytesIO(audio_data_bytes)

        audio_segment: AudioSegment = AudioSegment.from_raw(buffer, frame_rate=22050, channels=1, sample_width=2)

        output_bytes_io = io.BytesIO()
        audio_segment.export(output_bytes_io, format="wav")
        return self.create_synthesis_result_from_wav(
            file=output_bytes_io, message=message, chunk_size=chunk_size
        )



import io
import logging
from typing import Any, Optional
from pydub import AudioSegment

from vocode import getenv
from vocode.streaming.synthesizer.base_synthesizer import (
    BaseSynthesizer,
    SynthesisResult,
)
from vocode.streaming.models.synthesizer import ElevenLabsSynthesizerConfig
from vocode.streaming.agent.bot_sentiment_analyser import BotSentiment
from vocode.streaming.models.message import BaseMessage


ADAM_VOICE_ID = "pNInz6obpgDQGcFmaJgB"


class ElevenLabsSynthesizer(BaseSynthesizer):

    def __init__(
        self,
        config: ElevenLabsSynthesizerConfig,
        logger: Optional[logging.Logger] = None,
    ):
        super().__init__(config)
        
        import elevenlabs
        self.elevenlabs = elevenlabs
        
        self.api_key = config.api_key or getenv("ELEVEN_LABS_API_KEY")
        self.voice_id = config.voice_id or ADAM_VOICE_ID
        self.stability = config.stability
        self.similarity_boost = config.similarity_boost
        self.words_per_minute = 150
        

    def create_speech(
        self,
        message: BaseMessage,
        chunk_size: int,
        bot_sentiment: Optional[BotSentiment] = None,
    ) -> SynthesisResult:

        self.elevenlabs.set_api_key(self.api_key)
        voice = self.elevenlabs.Voice(voice_id=self.voice_id)
        if self.stability is not None and self.similarity_boost is not None:
            voice.settings = self.elevenlabs.VoiceSettings(
                stability=self.stability, similarity_boost=self.similarity_boost)

        audio = self.elevenlabs.generate(message.text, voice=voice)

        audio_segment: AudioSegment = AudioSegment.from_mp3(
            io.BytesIO(audio)
        )

        output_bytes_io = io.BytesIO()

        audio_segment.export(output_bytes_io, format="wav")

        return self.create_synthesis_result_from_wav(
            file=output_bytes_io,
            message=message,
            chunk_size=chunk_size,
        )
